# run with : python3 -m unittest TestTools

import unittest
import numpy as numpy
import math
from datetime import date

from BayesicFitting import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class TestTools( unittest.TestCase ) :
    """
    Test harness for Fitter class.

    @author Do Kester

    """
    def testtoday( self ):
        print( date.today() )

    def testMakeNext( self ):
        print( "===== makeNext ===========================" )
        x = 2
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( nx == 2 )
        xgen.close()

        x = [2,3,4,5]
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( ( nx == (k+2) and k < 4 ) or nx == 5 )
        xgen.close()

        x = [2,3,4,5,6,7,8,9,10,11]
        xgen = Tools.makeNext( x, 0 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( nx == (k+2) )

        x = [2,3,4,5]
        xgen = Tools.makeNext( x, 2 )
        for k in range( 10 ) :
            nx = next( xgen )
            print( k, nx )
            self.assertTrue( ( nx == (k+4) and k < 6 ) or nx == 5 )
        xgen.close()

    def testLength( self ) :
        print( "===== length ================================" )
        x = 3.0
        self.assertTrue( Tools.length( x ) == 1 )
        x = [2,3,4,5]
        self.assertTrue( Tools.length( x ) == 4 )
        x = numpy.asarray( [2,3,4,5] )
        self.assertTrue( Tools.length( x ) == 4 )
        x.resize( 2, 2 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( Tools.length( x ) == 2 )
        x = [2,3,4,5]
        x = numpy.asarray( [x] )
        print( x, x.shape, x.__class__ )
        self.assertTrue( Tools.length( x ) == 1 )

    def testToArray( self ) :
        print( "===== toArray ================================" )
        x = Tools.toArray( 3 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( 3.0 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4,5,6] )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4.5,5,6], ndim=1 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 1 )
        x = Tools.toArray( [3,4.5], ndim=2 )
        print( x, x.shape, x.__class__ )
        self.assertTrue( isinstance( x, numpy.ndarray ) and x.ndim == 2 )

    @classmethod
    def suite( cls ):
        return ConfiguredTestCase.suite( PriorTest.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestEvidence2

import unittest
import numpy
import math
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt

from FitPlot import plotFit
from FitPlot import plotErrdis
from FitPlot import plotErrdis2d


class TestEvidence2( unittest.TestCase  ) :

    def plot1( self ) :
        self.test1( plot=True )

    def plot2( self ) :
        self.test2( plot=True )

    def plot3( self ) :
        self.test3( plot=True )

    def plot4( self ) :
        self.test4( plot=True )

    def plot6( self ) :
        self.test6( plot=True )

    def plot7( self ) :
        self.test7( plot=True )

    def test1( self, plot=False ) :
        print( "====test1============================" )
        nn = 100
        x = numpy.zeros( nn, dtype=float )
        ym = 0.2 + 0.5 * x
        nf = 1.0
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-2,2]

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution ( x, y )

        logz1, maxll = plotErrdis( errdis, pm, limits=limits,
                                    max=0, plot=plot )

        print( "logZ  ", fmt( logz1 ) )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        stdv = ns.stdevs
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( stdv ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 2*dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo, numpy.exp( llevo ), 'r,' )

            mxl = numpy.exp( numpy.max( llevo ) ) * 1.2
            plt.plot( [pars,pars], [0.0,mxl], 'b-' )
            plt.plot( [par2,par2], [0.0,mxl], 'r-' )
            plt.plot( [par2,par2]+stdv, [0.0,mxl], 'g-' )
            plt.plot( [par2,par2]-stdv, [0.0,mxl], 'g-' )

            plt.show()

    def test2( self, plot=False ) :
        print( "====test2============================" )
        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.4 + 0.5 * x
        nf = 0.5
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution ( x, y )

        logz1, logl1 = plotErrdis2d( errdis, pm, limits=limits, max=0,
                    plot=plot )

        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 1.0 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.show()

    def test2_1( self, plot=False ) :
        print( "====test2_1============================" )
        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.2 + 0.5 * x
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution ( x, y )

        logz1, logl1 = plotErrdis2d( errdis, pm, limits=limits, max=0,
                    plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 1.0 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.show()

    def test3( self, plot=False ) :
        print( "====test3============================" )
        nn = 10
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm, fixedScale=0.5 )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        if plot :
            plt.figure( "model" )
            plotFit( x, data=y, model=pm, ftr=bf, truth=ym, show=False )

        errdis = GaussErrorDistribution ( x, y, scale=0.5 )

        logz1, logl1 = plotErrdis2d( errdis, pm, limits=limits, max=0,
                        plot=plot )

        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        dis = GaussErrorDistribution( x, y, scale=0.5 )
        ns = NestedSampler( x, model, y, distribution=dis, verbose=0 )

        logE = ns.sample()
        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

#        print( logz0 - logz1, logz0 - logz2 )
        self.assertTrue( abs( logz2 - logz0 ) < dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )

            plt.figure( "model" )			# grab again
            yfit = ns.modelfit
            err = samples.monteCarloError( x )
            plt.plot( x, yfit + err, 'b-' )
            plt.plot( x, yfit - err, 'b-' )

            plt.show()

    def test4( self, plot=False ) :
        print( "====test4============================" )

        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.4 + 0.0 * x
        nf = 0.5
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [0,1]
        nslim  = [0.1,1.0]

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        scale = bf.scale
        logz0 = bf.getLogZ( limits=limits, noiseLimits=nslim )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ), "  scale  ", fmt( scale ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        if plot :
            plt.figure( "model" )
            plotFit( x, data=y, model=pm, ftr=bf, truth=ym, show=False )

        errdis = GaussErrorDistribution ( x, y )

        logz1, logl1 = plotErrdis2d( errdis, pm, limits=limits, nslim=nslim,
                        plot=plot )
        if plot :
            plt.plot( pars[0], scale, 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        dis = GaussErrorDistribution( x, y )
        dis.setLimits( nslim )
        ns = NestedSampler( x, model, y, distribution=dis, verbose=0 )

        logE = ns.sample()
        par2 = ns.parameters
        scl2 = ns.scale
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ), "  scale  ", fmt( scl2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < 2 * dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        scevo =samples.getScaleEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], scevo, 'k,' )

            plt.figure( "model" )			# grab again
            yfit = ns.yfit
            err = samples.monteCarloError( x )
            plt.plot( x, yfit + err, 'b-' )
            plt.plot( x, yfit - err, 'b-' )

            plt.show()


    def test5( self, plot=None ) :
        print( "====test5============================" )
        nn = 10
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        s = 0.0
        s2 = 0.0
        mr = 10
        for k in range( mr ) :
            dis = GaussErrorDistribution( x, y, scale=0.5 )
            ns = NestedSampler( x, model, y, distribution=dis, verbose=0, seed=k )

            logE = ns.sample()
            par2 = ns.parameters
            logz2 = ns.logZ
            dlz2 = ns.logZprecision
            s += logz2
            s2 += logz2 * logz2
            print( "pars  ", fmt( par2 ), "  logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        logz = s / mr
        dlz = math.sqrt( s2 / mr - logz * logz )

        print( "Average  ", fmt( logz ), " +- ", fmt( dlz ) )

    def test6( self, plot=False ) :
        print( "====test6  Laplace ================" )

        nn = 20
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.9
        numpy.random.seed( 2345 )
        noise = numpy.random.laplace( size=nn )

        y = ym + nf * noise
        limits = [-1,2]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = AmoebaFitter( x, model, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = LaplaceErrorDistribution( x, y, scale=nf )
        logz1, logl1 = plotErrdis2d( errdis, model, limits=limits, max=0,
                plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, distribution='laplace', verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < 2 * dlz2 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )

            plt.show()

    def test6_0( self, plot=False ) :
        print( "====test6_0  Laplace ================" )

        nn = 20
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.0 * x
        nf = 0.9
        numpy.random.seed( 2345 )
        noise = numpy.random.laplace( size=nn )

        y = ym + nf * noise
        limits = [-1,2]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = PowellFitter( x, model, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", pars )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = LaplaceErrorDistribution( x, y, scale=nf )
        logz1, logl1 = plotErrdis( errdis, model, limits=limits, max=0,
                plot=plot )
        if plot :
            plt.plot( pars[0], logl1, 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )


        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, distribution='laplace', verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        stdv = ns.stdevs
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( stdv ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

#        self.assertTrue( abs( logz2 - logz0 ) < dlz2 )
        print( logz1 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo, numpy.exp( llevo ), 'r,' )

            mxl = numpy.exp( numpy.max( llevo ) ) * 1.2
            plt.plot( [pars,pars], [0.0,mxl], 'b-' )
            plt.plot( [par2,par2], [0.0,mxl], 'r-' )
            plt.plot( [par2,par2]+stdv, [0.0,mxl], 'g-' )
            plt.plot( [par2,par2]-stdv, [0.0,mxl], 'g-' )

            plt.show()

    def test6_1( self, plot=False ) :
        print( "====test6_1  Laplace ================" )

        nn = 20
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.9
        numpy.random.seed( 2345 )
        noise = numpy.random.laplace( size=nn )

        y = ym + nf * noise
        limits = [-1,2]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = AmoebaFitter( x, model, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = LaplaceErrorDistribution( x, y, scale=nf )
        logz1, logl1 = plotErrdis2d( errdis, model, limits=limits, max=0,
                        plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )


        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, distribution='laplace', verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 5 * dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )

            plt.show()

    def test7( self, plot=False ) :
        print( "====test7  Poisson ================" )

        nn = 100
        x = numpy.linspace( 0, 10, nn, dtype=float )
        ym = 1.9 + 2.2 * x
        numpy.random.seed( 2345 )
        y = numpy.random.poisson( ym, size=nn )

        limits = [0,4]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = AmoebaFitter( x, model, errdis="poisson" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = PoissonErrorDistribution( x, y )
        logz1, logl1 = plotErrdis2d( errdis, model, limits=limits, max=0,
                    plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = NestedSampler( x, model, y, distribution='poisson', verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        samples = ns.samples

        print( "pars  ", fmt( par2 ), fmt( samples.maxLikelihoodParameters ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < dlz2 )

        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )
            plt.show()


if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
from astropy import units
import math
from . import Tools

from .PolynomialModel import PolynomialModel
from .Dynamic import Dynamic
from .Prior import Prior
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018        Do Kester

class PolynomialDynamicModel( PolynomialModel, Dynamic ):
    """
    General polynomial model of an adaptable degree.

    f( x:p ) = &sum; p_k * x^k

    where the sum is over k running from 0 to degree ( inclusive ).

    It is a linear model.

    Author       Do Kester

    Examples
    --------
    >>> poly = PolynomialDynamicModel( )         # polynomial with unknown degree
    >>> ran = MoreRandom( )
    >>> poly.grow( )                         # starts at degree = 0, npar = 1
    >>> poly.grow( )                         # each grow( ) adds 1
    >>> poly.grow( )
    >>> poly.grow( )
    >>> print poly.npchain
    5
    >>> poly.shrink( )                        # shrink( ) deletes 1 degree
    >>> print poly.npbase
    4

    Category     mathematics/Fitting

    """

    deltaNpar = 1

    def __init__( self, degree, minDegree=0, maxDegree=None, fixed=None,
                  growPrior=None, copy=None, **kwargs ):
        """
        Polynomial of a adaptable degree.

        The model starts as a PolynomialModel of degree = 0.
        Growth of the model is governed by a exponential prior ( scale=1 ).

        Parameters
        ----------
        degree : int
            degree to start with; it should be minDegree <= degree <= maxDegree
        minDegree : int
            minimum degree of polynomial (def=0)
        maxDegree : None or int
            maximum degree of polynomial (def=None)
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxDegree is None else UniformPrior
        copy : PolynomialDynamicModel
            model to copy

        Raises
        ------
        AttributeError when fixed parameters are requested
        ValueError when degree is outside [min..max] range
        """
        if fixed is not None :
            raise AttributeError( "DynamicModel cannot have fixed parameters" )
        if degree < minDegree or ( maxDegree is not None and degree > maxDegree ):
            raise ValueError( "degree outside range of [min..max] range" )


        super( PolynomialDynamicModel, self ).__init__( degree, copy=copy, **kwargs )

        if copy is None :
            self.minComp = minDegree + 1
            self.maxComp = maxDegree if maxDegree is None else maxDegree + 1
            if growPrior is None :
                if maxDegree is None :
                    self.growPrior = ExponentialPrior( scale=2 )
                else :
                    lim = [minDegree+1, maxDegree+2]    # limits on components
                    self.growPrior =  UniformPrior( limits=lim )
            else :
                self.growPrior = growPrior
        else :
            self.minComp = copy.minComp
            self.maxComp = copy.maxComp
            self.growPrior = copy.growPrior.copy()

    def copy( self ):
        """ Copy method.  """
        return PolynomialDynamicModel( self.degree, copy=self )

    def changeNComp( self, dn ) :
        self.degree += dn

    def __setattr__( self, name, value ) :
        lnon = {"maxComp": int}
        dind = {"degree": int, "minComp": int, "maxComp": int,
                "growPrior": Prior}
        if ( Tools.setNoneAttributes( self, name, value, lnon ) or
             Tools.setSingleAttributes( self, name, value, dind ) ) :
            pass
        else :
            super( ).__setattr__( name, value )


    def baseName( self ):
        """ Return a string representation of the model.  """
        return "Dynamic" + super( PolynomialDynamicModel, self ).baseName( )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class PowerLawModel( NonLinearModel ):
    """
    General powerlaw model of arbitrary degree.
    .. math::
        f( x:p ) = p_0 * ( x - p_1 )^p_2

    with
        p_0 = amplitide
        p_1 = x-shift
        p_2 = power

    The parameters are initialized at {1.0, 0.0, 1.0}.

    Note that the term ( x - p_1 ) needs to be divided by a factor 1.0
    in the same units as the x, to get the overall units of f( x:p ) right.
    The factor is omitted as it does not contribute in the calculations.

    Examples
    --------
    >>> pl = PowerLawModel( )
    >>> print( pl.npchain )
    4

    """
    def __init__( self, copy=None, **kwargs ):
        """
        Powerlaw of an unknown degree.

        The number of parameters is 3

        Parameters
        ----------
        copy : PowerLawModel
            to be copied
        """
        param = [1.0, 0.0, 1.0]
        names = ["amplitude", "x-shift", "power"]

        super( PowerLawModel, self ).__init__( 3, copy=copy, params=param,
                    names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return PowerLawModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return params[0] * numpy.power( xdata - params[1], params[2] )


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input (xdata) value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        partial = numpy.ndarray( ( Tools.length( xdata ), self.npbase ) )
        p0 = params[0]
        xs = xdata - params[1]
        p2 = params[2]

        parts = { 0 : ( lambda: numpy.power( xs, p2 ) ),
                  1 : ( lambda: -p0 * p2 * numpy.power( xs, p2 - 1 ) ),
                  2 : ( lambda: p0 * numpy.power( xs, p2 ) * numpy.log( xs ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial


    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative (df/dx) at the input (xdata) value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return params[0] * params[2] * numpy.power( xdata - params[1], params[2] - 1 )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "PowerLaw: f( x:p ) = p_0 * ( x - p_1)^p_2" )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        if k == 1:
            return self.xUnit
        return units.Unit( 1.0 )





#run with: python3 -m unittest TestCompoundModel

import unittest
import numpy as numpy
from astropy import units
import math

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class TestDynamicModel( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """

    def test1Model1( self ):
        print( "  Test PolynomialDynamicModel" )
        m = PolynomialDynamicModel( 0 )
        self.growshrink( m, dnp=1 )

    def test1Model2( self ):
        print( "  Test HarmonicDynamicModel" )
        m = HarmonicDynamicModel( 1 )
        self.growshrink( m, dnp=2 )

    def test1Model3( self ):
        print( "  Test RepeatingModel" )
        m = GaussModel()
        r = RepeatingModel( 1, m  )
        self.growshrink( r, dnp=3 )


    def growshrink( self, m, dnp=1 ) :
        self.assertTrue( m.npchain == dnp )
        self.assertTrue( m.npbase == dnp )

        Tools.printclass( m )

        m.grow()
        m.grow()
        m.grow()

        print( m.npchain, m.npbase, m.npmax )
        self.assertTrue( m.npchain == 4 * dnp )
        self.assertTrue( m.npbase == 4 * dnp )

        m.grow()
        m.shrink()

        Tools.printclass( m )
        self.assertTrue( m.npchain == 4 * dnp )
        self.assertTrue( m.npbase == 4 * dnp )

        m.shrink()
        self.assertTrue( m.npchain == 3 * dnp )
        self.assertTrue( m.npbase == 3 * dnp )


    def test2Model( self ):
        print( "========================" )
        print( "  Test two models" )
        print( "========================" )

        self.assertRaises( ValueError, PolynomialDynamicModel, 0, minDegree=2, maxDegree=4 )

        m = GaussModel( )
        p = PolynomialDynamicModel( 2, minDegree=2, maxDegree=4 )
        print( p.growPrior )
        m += p

        self.assertTrue( m.isDynamic() )
        self.assertFalse( isinstance( m, Dynamic ) )
        self.assertTrue( isinstance( m._next, Dynamic ) )

        lolim = [-10, -1, 0, -10]
        hilim = [+10, +1, 10, 10]
        m.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( m )
        Tools.printclass( p )

        par = numpy.asarray( [1,-0.2,0.3,-1.2,0.2,0.1] )
        x = numpy.linspace( -1.0, 1.0, 11 )
        y = numpy.linspace( 0.0, 2.3, 11 )

        errdis = GaussErrorDistribution( x, y )
        sl = SampleList( m, 10, errdis )
        print( m. npchain, m.npbase, m._next.npbase, hex(id(m)) )
        print( "====================" )

        seng = StartEngine( sl, errdis )
        for s in sl :
            print( s.model.npchain, s.model.npbase, s.model._next.npbase, hex(id(s.model)) )
            seng.execute( s, 0.0 )

        logl = sl.getLogLikelihoodEvolution()
        lowl = numpy.min( logl )
        print( "logl  ", fmt( logl, max=None ) )
        print( "lowl  ", fmt( lowl ) )

        beng = BirthEngine( sl, errdis )
        deng = DeathEngine( sl, errdis )
        print( "start  np ", m.npchain )
        for s in sl :
            Tools.printclass( s )
            suc = beng.execute( s, lowl )
            print( "Birth  :", suc, s.model.npchain, s.model.getLimits() )
            suc = deng.execute( s, lowl )
            print( "Death  :", suc, s.model.npchain, s.model.getLimits() )


    def testNestedSampler( self, plot=False ) :
        print( "========================" )
        print( "  Test Nested Sampler" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 21
        x = numpy.linspace( -2.0, 3.0, N, dtype=float )
        y = 1.2 + 0.5 * x + 0.33 * x * x + 0.27 * x * x * x
        noise = numpy.random.randn( N )
        y += 0.02 * noise

        pm = PolynomialDynamicModel( 0 )
        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        engines = None
        ns = NestedSampler( x, pm, y, seed=2031967, engines=engines )
        ns.verbose = 2

        ns.distribution.setLimits( [0.01, 100] )

        logE = ns.sample( plot=plot )

    def plotNestedSampler2( self ) :
        self.testNestedSampler2( plot=True )

    def testNestedSampler2( self, plot=False ) :
        print( "========================" )
        print( "  Test Nested Sampler 2" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 40
        x = numpy.arange( N, dtype=float ) / 10
        y = [0, 1, 3, 4, 6, 7, 5, 8, 9, 2, -2,-9,-8,-5,-7,-6,-4,-3,-1, 0] * 2
        y - numpy.asarray( y, dtype=float )
#        y = numpy.linspace( 0.0, 4.0, N, dtype=float ) % 2
        noise = numpy.random.randn( N )
        y += 0.3 * noise

        pm = HarmonicDynamicModel( 1, period=2.0 )
        pm.growPrior.scale = 6
        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ns = NestedSampler( x, pm, y, seed=2031967 )
        ns.verbose = 2
        ns.distribution.setLimits( [0.01, 100] )
#        ns.distribution.scale = 0.02

        logE = ns.sample( plot=plot )
        print( "scale  ", ns.scale )

    def plotNestedSampler3( self ) :
        self.testNestedSampler3( plot=True )

    def testNestedSampler3( self, plot=False ) :
        print( "========================" )
        print( "  Test Nested Sampler 3" )
        print( "========================" )

        numpy.random.seed( 1315 )
        N = 200
        x = numpy.arange( N, dtype=float ) / 40
        y = 0.3 * numpy.random.randn( N )
        gm = GaussModel()
        for k in range( 1, 6 ) :
            par = [5.0/k, k-0.5, 0.1]
            y += gm.result( x, par )

        pm = RepeatingModel( 1, gm )
        pm.growPrior.scale = 5
        lolim = [0.0,0.0,0.0]
        hilim = [+10,5.0,1.0]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ns = NestedSampler( x, pm, y, seed=2031967 )
        ns.verbose = 2
        ns.distribution.setLimits( [0.01, 100] )
#        ns.distribution.scale = 0.02

        logE = ns.sample( plot=plot )
        sl = ns.samples
        print( "maxlik   ", fmt( sl.maxLikelihoodParameters, max=None ) )
        print( "median   ", fmt( sl.medianParameters, max=None ) )
        print( "modus    ", fmt( sl.modusParameters, max=None ) )
        print( "scale    ", fmt( ns.scale ) )


    def suite( cls ):
        return unittest.TestCase.suite( TestDynamicModel.__class__ )


if __name__ == '__main__':
    unittest.main()




import numpy as numpy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .NoiseScale import NoiseScale

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class LaplaceErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Laplace likelihood.

    For one residual, x, it holds
    .. math::
        f( x ) = 1 / ( 2 s ) exp( - |x| / s )

    where s is the scale.
    s is a hyperparameter, which might be estimated from the data.

    The variance of this function is :math:`\sigma^2 = 2 s ^ 2`.
    See: toSigma()

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.
    .. math::
        logL = log( N / ( 2 s ) ) - \sum( |x| / s  )

    Using weights this becomes:
    .. math::
        logL = log( \sum( w ) / ( 2 s ) ) - \sum( w |x| / s  )

    Using this error distribution results in median-like solutions.

    Author       Do Kester.

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2.0 )


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, scale=1.0, limits=None,
                  copy=None ) :
        """
        Constructor of Laplace Distribution.

        Parameters
        ----------
        xdata : array_like
            input data for the model
        data : array_like
            data to be fitted
        weights : array_like
            weights to be used
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.

        copy : LaplaceErrorDistribution
            distribution to be copied.
        """
        super( LaplaceErrorDistribution, self ).__init__( xdata, data,
               weights=weights, scale=scale, limits=limits, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return LaplaceErrorDistribution( self.xdata, self.data, copy=self )

    #  *********DATA & WEIGHT***************************************************
    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True

    def toSigma( self, scale ) :
        """
        Return sigma, the squareroot of the variance.
        Parameter
        --------
        scale : float
            the scale of this Laplace distribution.
        """
        return scale * math.sqrt( 2.0 )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood( self, model, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1
        np = model.npchain
        scale = allpars[-1]
        res = self.getResiduals( model, allpars[:np] )
        sumres = self.getSumRes( res, scale )
        return - self.sumweight * ( self.LOG2 + math.log( scale ) ) - sumres

    def getScale( self, model ) :
        """
        Return the noise scale

        Parameters
        ----------
        model : Model
            the model involved
        """
        sumres = self.getSumRes( self.getResiduals( model ), 1.0 )
        return sumres / self.sumweight


    def getSumRes( self, residual, scale ):
        """
        Return the sum of the absolute values of the normalized residuals.

        ..math ::
            \sum ( | res | / scale )

        Parameters
        ----------
        residual : array_like
            the residuals
        scale : float
            the noise scale

        """
        if self.weights is not None :
            residual = residual * self.weights
        return numpy.sum( numpy.abs( residual ) ) / scale

    def partialLogL( self, model, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1
        np = model.npchain
        scale = allpars[-1]

        dM = model.partial( self.xdata, allpars[:np] )
        dL = numpy.zeros( len( fitIndex ), dtype=float )
        res = self.getResiduals( model, allpars[:np] )
        wgt = numpy.ones_like( res, dtype=float ) if self.weights is None else self.weights
        wgt = numpy.copysign( wgt, res )

        i = 0
        for k in fitIndex :
            if k >= 0 :
                dL[i] = numpy.sum( wgt * dM[:,k] )
            else :
                dL[i] = self.getSumRes( res, scale ) - self.sumweight
            i += 1
        return dL / scale

    def __str__( self ) :
        return "Laplace error distribution"




# run with : python3 -m unittest TestRobustShell

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal as assertAAE

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2004 SRON, Do Kester

class TestRobustShell( unittest.TestCase ):
    """
    Test harness for RobustShell.

    Author       Do Kester

    """
    def plot1( self ) :
        self.test1( plot=True )
        plt.show()

    def plot2( self ) :
        self.test2( plot=True )
        plt.show()

    #  **************************************************************
    def test1( self, plot=False ):
        """     #   test slope fit  """

        print( "\n   Robust fitter Test 1  \n" )

        ndata = 101
        aa = 5.0            #  average offset
        bb = 2.0            #  slope
        ss = 0.3            #  noise Normal distr.

        x = numpy.linspace( -1, +1, ndata, dtype=float )
        numpy.random.seed( 12345 )
        y = aa + bb * x + ss * numpy.random.randn( ndata )

        model = PolynomialModel( 1 )
        fitter = Fitter( x, model )

        print( "Testing Straight Line fit" )

        par = fitter.fit( y )
        std = fitter.stdevs
        chisq = fitter.chisq
        print( "truth   " + fmt( aa ) + fmt( bb ) )
        print( "params  " + fmt( par ) )
        print( "stdevs  " + fmt( std ) )
        print( "chisq   " + fmt( chisq ) )
        assertAAE( par, numpy.asarray( [aa, bb] ), 2 )

        if plot :
            plt.plot( x, y, 'k.' )
            plt.plot( x, model( x ), 'k-' )

        # make some outliers
        ko = [10*k+4 for k in range( 10 )]
        ko = numpy.asarray( ko )
        y[ko] += numpy.linspace( -5, 2, 10 )

        romod = PolynomialModel( 1 )
        altfit = Fitter( x, romod )
        alt = altfit.fit( y )
        ast = altfit.stdevs
        altch = altfit.chisq
        print( "params  " + fmt( alt ) )
        print( "stdevs  " + fmt( ast ) )
        print( "chisq   " + fmt( altch ) )

        if plot :
            plt.plot( x, romod( x ), 'b-' )

        rf = RobustShell( altfit )
        Tools.printclass( rf )

        alt = rf.fit( y )
        ast = altfit.stdevs
        altch = altfit.chisq
        print( rf )
        print( "params  " + fmt( alt ) )
        print( "stdevs  " + fmt( ast ) )
        print( "chisq   " + fmt( altch ) + fmt( rf.scale ) )
        print( "weight  " + fmt( rf.weights[ko], max=None ) )
        if plot :
            plt.plot( x, romod( x ), 'g-' )
            plt.plot( x, rf.weights, 'g-' )
        assertAAE( par, alt, 1 )
        assertAAE( std, ast, 1 )

        rf = RobustShell( altfit, kernel=Cosine(), domain=4.0 )
        alt = rf.fit( y )
        ast = altfit.stdevs
        altch = altfit.chisq
        print( rf )
        print( "params  " + fmt( alt ) )
        print( "stdevs  " + fmt( ast ) )
        print( "chisq   " + fmt( altch ) + fmt( rf.scale ) )
        print( "weight  " + fmt( rf.weights[ko], max=None ) )
        if plot :
            plt.plot( x, romod( x ), 'r-' )
            plt.plot( x, rf.weights, 'r-' )
        assertAAE( par, alt, 1 )
        assertAAE( std, ast, 1 )

        rf = RobustShell( altfit, kernel=Huber, domain=1.0 )
#        rf.setNoiseScale( 0.3 )
        alt = rf.fit( y )
        ast = altfit.stdevs
        altch = altfit.chisq
        print( rf )
        print( "params  " + fmt( alt ) )
        print( "stdevs  " + fmt( ast ) )
        print( "chisq   " + fmt( altch ) + fmt( rf.scale ) )
        print( "weight  " + fmt( rf.weights[ko], max=None ) )
        if plot :
            plt.plot( x, romod( x ), 'c-' )
            plt.plot( x, rf.weights, 'c-' )
        assertAAE( par, alt, 1 )
        assertAAE( std, ast, 1 )

        rf = RobustShell( altfit, kernel=Uniform )
        alt = rf.fit( y )
        ast = altfit.stdevs
        altch = altfit.chisq
        print( rf )
        print( "params  " + fmt( alt ) )
        print( "stdevs  " + fmt( ast ) )
        print( "chisq   " + fmt( altch ) + fmt( rf.scale ) )
        print( "weight  " + fmt( rf.weights[ko], max=None ) )
        if plot :
            plt.plot( x, romod( x ), 'm-' )
            plt.plot( x, rf.weights, 'm-' )
        assertAAE( par, alt, 1 )
        assertAAE( std, ast, 1 )

    def test2( self, plot=False ):

        ndata = 101
        c0 = 3.2
        c1 = -0.1
        c2 = 0.3
        c3 = 1.1
        c4 = 2.1
        ss = 0.2

        x = numpy.linspace( -1, +1, ndata, dtype=float )
        y = ( x - c1 ) / c2
        numpy.random.seed( 12345 )
        y = c0 * numpy.exp( -0.5 * y * y ) + x * c3 + c4 + ss * numpy.random.randn( ndata )

        print( "Testing Nonlinear Fitters with RobustShell." )

        print( fmt( [c0,c1,c2,c4,c3] ) )

        modl2 = GaussModel( )
        modl2 += PolynomialModel( 1 )

#        lmfit = CurveFitter( x, modl2 )
        lmfit = LevenbergMarquardtFitter( x, modl2 )
        par2 = lmfit.fit( y )
        std2 = lmfit.stdevs
        print( fmt( par2, max=5 ) )
        print( fmt( std2, max=5 ) )

        # make some outliers
        ko = [10*k+4 for k in range( 10 )]
        ko = numpy.asarray( ko )
        y[ko] += numpy.linspace( -5, 2, 10 )

        if plot :
            plt.plot( x, y, 'k.' )
            plt.plot( x, modl2( x ), 'k-' )


#        lmfit = CurveFitter( x, modl2 )
        lmfit = LevenbergMarquardtFitter( x, modl2 )
#        lmfit.setParameters( initpar2 )
        rf = RobustShell( lmfit )
#        rf.setVerbose( 1 )
        print( str( rf ) )
        par1 = rf.fit( y )
        std1 = rf.stdevs
        print( fmt( par1, max=5 ) )
        print( fmt( std1, max=5 ) )
        print( fmt( par2, max=5 ) )
        print( fmt( rf.weights[ko], max=20 ) )
        if plot :
            plt.plot( x, modl2( x ), 'g-' )
            plt.plot( x, rf.weights, 'g-' )
        assertAAE( par2, par1, 1 )
        assertAAE( std2, std1, 1 )



#        lmfit = CurveFitter( x, modl2 )
        lmfit = LevenbergMarquardtFitter( x, modl2 )
#        lmfit.setParameters( initpar2 )
        rf = RobustShell( lmfit, onesided="negative" )
#        rf.setVerbose( 1 )
        print( rf )
        par3 = rf.fit( y )
        std3 = rf.stdevs
        print( fmt( par3, max=5 ) )
        print( fmt( std3, max=5 ) )
        print( fmt( par2, max=5 ) )
        print( fmt( rf.weights[ko], max=20 ) )

        if plot :
            plt.plot( x, modl2( x ), 'r-' )
            plt.plot( x, rf.weights, 'r-' )
        assertAAE( par2, par3, 1 )
        assertAAE( std2, std3, 1 )


    def suite( cls ):
        return unittest.TestCase.suite( TestRobustShell.__class__ )

if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
import math
from . import Tools
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2011 Do Kester, SRON (JAVA code)
#  *    2016 - 2017 Do Kester

class LorentzModel( NonLinearModel ):
    """
    Lorentzian Model.
    .. math::
        f( x:p ) = p_0 * ( p_2^2 / ( ( x - p_1 )^2 + p_2^2 )

    :math:`p_0` = amplitude
    :math:`p_1` = x-shift
    :math:`p_2` = gamma ( width )

    The parameters are initialized at [1/PI, 0.0, 1.0] where the integral
    over the function equals 1.
    Parameter 2 ( gamma ) is always kept stricktly positive ( >0 ).

    This model is also known as Cauchy or Cauchy-Lorentz.

    Notes
    -----
    There are other possible
    <a href="http://en.wikipedia.org/wiki/Cauchy_distribution#Probability_density_function">
    definitions of this model</a>, where the integral equals 1.0.
    Definitions that integrate to 1.0 are more fit as a distribution function.
    See sample/CauchyErrorDistribution.

    We choose our definition for 2 reasons.
    1. to be in line with the definitions of the GaussModel, SincModel,
        VoigtModel, all KernelModels etc. In all of them the amplitude parameter,
        :math:`p_0`, equals the maximum of the function.
        I.e. :math:`p_0` is indeed the amplitude.
    2. to have maximally independent parameters, meaning that if you change one
        parameter, only that aspect changes. In the present definition this is the case.
        In the alternative definition if you change :math:`p_2`, not only the width
        of the function changes, but also the amplitude.

    Examples
    --------
    >>> lorentz = LorentzModel( )
    >>> lorentz.setParameters( [5, 4, 1] )
    >>> print( lorentz( numpy.arange(  41 , dtype=float ) / 5 ) )

    """
    def __init__( self, copy=None, **kwargs ):
        """
        Lorentzian model.

        Number of parameters is 3.

        Parameters
        ----------
        copy : LorentzModel
            to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        param = [1/math.pi, 0.0, 1.0]
        names = ["amplitude", "center", "width"]

        super( LorentzModel, self ).__init__( 3, copy=copy, params=param,
                        names=names, **kwargs )

        if copy is None :
            self.posIndex = [2]
            self.nonZero = [2]


    def copy( self ):
        """ Copy method.  """
        return LorentzModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        s2 = params[2] * params[2]
        x = ( xdata - params[1] )
        result = params[0] * s2 / ( x * x + s2 )
        return result

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        partial = numpy.ndarray( (Tools.length( xdata ), self.npbase ) )

        a = params[0]
        s = abs( params[2] )
        s2 = s * s
        x = ( xdata - params[1] )
        e = 1 / ( x * x + s2 )

        parts = { 0 : ( lambda: s2 * e ),
                  1 : ( lambda: a * 2 * s2 * x * e * e ),
                  2 : ( lambda: a * 2 * s * x * x * e * e ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial


    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the derivative
        params : array_like
            values for the parameters.

        """
        x = xdata - params[1]
        c2 = params[2] * params[2]

        return -2 * params[0] * c2 * x / numpy.square( x * x + c2 )


    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "Lorentz: f( x:p ) = p_0 * p_2^2 / ( ( x - p_1 )^2 + p_2^2 )" )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit





# run with : python3 -m unittest TestLogFactorial

import unittest
import numpy as numpy
import math
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import matplotlib.pyplot as plt
import warnings

from BayesicFitting import logFactorial

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestLogFactorial( unittest.TestCase ):
    """
    Test harness for LogFactorial

    Author:      Do Kester

    """
    def testLogFactorial( self, plot=False ):

        print( "====LogFactorial ===================" )
        self.assertTrue( logFactorial( 0 ) == 0 )
        self.assertTrue( logFactorial( 1 ) == 0 )
        self.assertTrue( logFactorial( 3 ) == 1.7917594692280550 )
        self.assertTrue( logFactorial( 10 ) == 15.1044125730755159 )
        print( logFactorial( 99 ), logFactorial( 100 ), logFactorial( 101 ) )
        self.assertTrue( logFactorial( 99 ) == 359.1342053695754544 )
        lf = 359.1342053695754544 + math.log( 100 )
        assertAAE( logFactorial( 100 ), lf )

        kk = [3, 5, 7]
        print( logFactorial( kk ) )
        self.assertTrue( isinstance( kk, list ) )

        data = numpy.arange( 100, dtype=int ) + 40

        logd = numpy.log( data )

        lf1 = numpy.cumsum( logd ) + logFactorial( 39 )
        lf2 = logFactorial( data )

        k = [0,10,20,30,40,50,60,70,80,90]
        print( lf1[k] )
        print( lf2[k] )
        assertAAE( lf1, lf2 )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestLogFactorial.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestEngine2

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestEngine2( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestEngine2, self ).__init__( name )


    def initEngine( self, order=1, np=101 ):
        m = PolynomialModel( order )

        up1 = UniformPrior( limits=[-10,10])
        up2 = UniformPrior( limits=[  0,10] )

        m.priors = [up1, up2]

        xdata = numpy.linspace( 0.0, 10.0, np )
        data = numpy.ceil( numpy.arange( np, dtype=float ) / 2.3 + 2.4 )
        numpy.random.seed( 345345 )
        data += 0.4 * numpy.random.randn( np )
#        print( data )
        numpy.set_printoptions( precision=3, suppress=True )
        return ( m, xdata, data )

    def plotall( self ) :
        self.testRandomEngine( plot=True )
        self.testGibbsEngine( plot=True )
        self.testStepEngine( plot=True )
        self.testGalileanEngine( plot=True )


    def plotRandomEngine( self ):
        self.testRandomEngine( plot=True )

    def testRandomEngine( self, plot=False ):
        print( "\n   Random Engine Test\n" )
        self.stdenginetest( RandomEngine, iter=200, nsamp=10, plot=plot )

    def plotGibbsEngine( self ):
        self.testGibbsEngine( plot=True )

    def testGibbsEngine( self, plot=False ):
        print( "\n   Gibbs Engine Test\n" )
        self.stdenginetest( GibbsEngine, iter=101, nsamp=10, plot=plot )

    def plotStepEngine( self ):
        self.testStepEngine( plot=True )

    def testStepEngine( self, plot=False ):
        print( "\n   Step Engine Test\n" )
        self.stdenginetest( StepEngine, iter=1001, nsamp=10, plot=plot )

#    def plotCrossEngine( self ):
#        self.testCrossEngine( plot=True )

#    def testCrossEngine( self, plot=False ):
#        print( "\n   Cross Engine Test\n" )
#        self.stdenginetest( CrossEngine, iter=800, nsamp=100, plot=plot )

    def plotGalileanEngine( self ):
        self.testGalileanEngine( plot=True )

    def testGalileanEngine( self, plot=False ):
        print( "\n   Galilean Engine Test\n" )
        self.stdenginetest( GalileanEngine, iter=101, nsamp=10, plot=plot )

    def stdenginetest( self, myengine, nsamp=4, iter=100, plot=False ) :
        m, xdata, data = self.initEngine()
        errdis = GaussErrorDistribution( xdata, data, scale=0.5 )
        Tools.printclass( errdis )

        sl = SampleList( m, nsamp, errdis )
        map = numpy.ndarray( (41,21), dtype=float )
        mmx = -math.inf
        for k0 in range( 21 ) :
            for k1 in range( 41 ) :
                pl = [k0-10, 0.25*k1, 0.5]
                map[k1,k0] = errdis.logLikelihood( m, pl )
                if map[k1,k0] > mmx :
                    mmx = map[k1,k0]
                    mk1 = k1 / 4
                    mk0 = k0 - 10

#        print( map )
        print( mk0, mk1, mmx )

        ax = numpy.linspace( -10, 10, 21 )
        ay = numpy.linspace(   0, 10, 41 )
        v = [-500000,-100000,-50000,-10000,-5000, -3000, -1000, -500, -300, -200]

        engine = StartEngine( sl, errdis )
        Tools.printclass( engine.walkers[0] )
        for samp in engine.walkers :
            engine.execute( samp, -math.inf )

        pevo = sl.getParameterEvolution()

        if plot :
            plt.figure( str( myengine ) )
            plt.contour( ax, ay, map, v )
            plt.plot( [mk0], [mk1], 'k*' )
            plt.plot( pevo[:,0], pevo[:,1], 'k.' )

        col = ['k-', 'r-', 'g-', 'b-']
        engine = myengine( sl, errdis )
        for k in range( iter ) :
            lowL, klo = sl.getLowLogL()
            p0 = sl[klo].allpars[:2]
            engine.calculateUnitRange()
            while True :
                kok = engine.rng.randint( 0, nsamp )
#                print( klo, kok )
                if kok != klo :
                    sl[klo].allpars = sl[kok].allpars.copy()
                    break
            engine.execute( sl[klo], lowL )
            p1 = sl[klo].allpars[:2]
            if k % 50 == 0 :
                print( k, klo, p1, sl[klo].logL )

            if plot :
                plt.plot( [p0[0],p1[0]], [p0[1],p1[1]], col[klo%4] )

        print( lowL, p1 )
        print( mmx, mk0, mk1 )

#        engine.printReport()
        if plot :
            pevo = sl.getParameterEvolution()
            plt.plot( pevo[:,0], pevo[:,1], 'r.' )
            plt.show()

        self.assertTrue( lowL > mmx )
        self.assertTrue( abs( p1[0] - mk0 ) < 1 )
        self.assertTrue( abs( p1[1] - mk1 ) < 1 )


    def suite( cls ):
        return unittest.TestCase.suite( TestEngine2.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
import math

from .BaseFitter import BaseFitter

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2017 Do Kester

class QRFitter( BaseFitter ):
    """
    Fitter for linear models, using QR decomposition.
    The QRFitter class is to be used in conjunction with Model classes, linear
    in its parameters.

    For Linear models the matrix equation
    .. math::
    H * p = \beta

    is solved for p. H is the Hessian matrix ( :math:`D * w * D^T` )
    and :math:`\beta` is the inproduct of the data with the D, design matrix.
    .. math::
    \beta = y * w * D^T

    The QRFitter class use QR decomposition which effectively is an inversion
    of the hessian matrix such that
    .. math::
    p = \beta * H^{-1}

    It can be more efficient if
    similar ydata needs to be fitter to the same model and xdata.
    In that case it uses the same decomposition for all fits.

    Examples
    --------
    # assume x and y are numpy.asarray data arrays:
    >>> x = numpy.asarray.range( 100 )
    >>> poly = PolynomialModel( 1 )                             # line
    >>> fitter = QRFitter( x, poly )
    >>> for k in range( 1, 4 ) :
    >>>     y = numpy.arange( 100 ) // k                        # digitization noise
    >>>     param = fitter.fit( y )                             # use same QR decomposition
    >>>     stdev = fitter.stdevs                               # stdevs on the parameters
    >>>     print( k, param )
    >>>     print( " ", stdev )

    Category:    Mathematics/Fitting

    Attributes
    ----------
    needsNewDecomposition : bool
        True when starting. Thereafter False,
            i.e. the previous QR-decomposition is used, unless weights are used.
            Only for linear fitters, setting it to false might improve efficiency.
            It only works properly when the model, the x-data etc are exactly the
            same in the previous run.<br>
            Whenever weights are used, it always needs a new decomposition.

    qrmat : matrix
        matrix formed by :math:`q * r^{-1}`, where q,r is the QR decomposition
        of the deging matrix.
        qrmat is to be multiplied with the data vector to get the solution.

    """

    def __init__( self, xdata, model, map=False, keep=None ):
        """
        Create a new Fitter, providing xdatas and model.

        A Fitter class is defined by its model and the input vector (the
        independent variable). When a fit to another model and/or another
        input vector is needed a new object should be created.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
        map : bool (False)
            When true, the xdata should be interpreted as a map.
            The fitting is done on the pixel indices of the map,
            using ImageAssistant
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep will be used by the Fitter as long as the Fitter exists.
            See also `fit( ..., keep=dict )`

        """
        super( QRFitter, self ).__init__( xdata, model, map=map, keep=keep )

        self.needsNewDecomposition = True
        self.qrmat = None

    def fit( self, ydata, weights=None, keep=None ):
        """
        Return model parameters fitted to the data, including weights.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data ( = 1.0 / sigma^2 )
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.
        Raises
        ------
            ValueError when ydata or weights contain a NaN

        """
        fi, ydata, weights = self.fitprolog( ydata, weights=weights, keep=keep )

        if self.model.isNullModel() :
            self.chiSquared( ydata, weights )
            return numpy.asarray( 0 )

        ydatacopy = ydata.copy( )
        # subtract influence of fixed parameters on the data
        if fi is not None :
            fxpar = numpy.copy( self.model.parameters )
            fxpar[fi] = 0.0
            ydatacopy = numpy.subtract( ydatacopy, self.model.result( self.xdata, fxpar) )

        wgts = ( numpy.ones_like( ydata, dtype=float ) if weights is None else
                 numpy.sqrt( weights ) )

        if hasattr( self, "normdata" ) :
            normweight = math.sqrt( self.normweight )
            ydatacopy = numpy.append( ydatacopy, self.normdata )
            wgts = numpy.append( wgts, normweight )

        ydatacopy = ydatacopy * wgts

        if self.needsNewDecomposition or weights is not None:
            design = ( self.getDesign( index=fi ).transpose() * wgts ).transpose()

            # The QR decomposition in numpy has a different interface from that in scipy.
            # Here we use numpy. See:
            #   https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.qr.html

            q, r = numpy.linalg.qr( design )
            self.qrmat = numpy.dot( numpy.linalg.inv( r ), q.transpose() )
            self.needsNewDecomposition = False

        params = numpy.dot( self.qrmat, ydatacopy )

        params = self.insertParameters( params, index=fi )
        self.model.parameters = params
        self.chiSquared( ydata, weights=weights )

        return params

#  *************************************************************************
    def __str__( self ):
        """ Return the name of the fitter. """
        return "QRFitter"





import numpy as numpy
from . import Tools
from .Model import Model
from .LinearModel import LinearModel
from .SplinesModel import SplinesModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class SurfaceSplinesModel( LinearModel ):
    """
    Surface splines model of arbitrary order and knot settings.

    It is a linear model.

    Surface splines are direct product of a splines model in the x-direction with
    a splines model in the y-direction.

    The number of parameters is
    ( xknotlength + xorder - 1 ) * ( yknotlength + yorder - 1 )

    The SplinesModel has more information about order and knots.

    Examples
    --------
    >>> nxk = 17
    >>> nyk = 11
    >>> xknots = numpy.arange(  nxk , dtype=float ) * 10      # make knots from 0 to 160
    >>> yknots = numpy.arange(  nyk , dtype=float ) * 10      # make knots from 0 to 100
    >>> csm = SurfaceSplinesModel( xknots, yknots, 2 )
    >>> print csm.getNumberOfParameters( )      # ( nxk + order - 1 )*( nyk + order - 1 )
    216
    # ... fitter etc. see Fitter

    Category     mathematics/Fitting

    Attributes
    ----------

    """

    def __init__( self, knots, order=3, copy=None, fixed=None, **kwargs ):
        """
        Splines on a given set of knots and a given order.

        The number of parameters is ( length( knots ) + order - 1 )

        Parameters
        ----------
        knots : list of array_like
            positions of the knots in all dimensions
        order : int or list of ints
            order of the splines in all dimensions
        copy : SurfaceSplinesModel
            model to be copied
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        AttributeErrr : When fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "SurfaceSplinesModel cannot have fixed parameters" )

        super( SurfaceSplinesModel, self ).__init__( self.calcNp( knots, order ),
                    ndim=len( knots ), copy=copy )
        self.order = order if Tools.length( order ) > 1 else [order] * len( knots )
        self.knots = knots
        models = []
        for (kn,dr) in zip( self.knots, self.order ) :
            models = models + [SplinesModel( knots=kn, order=dr )]
        self.models = models

    def copy( self ):
        """ Copy method.  """
        return SurfaceSplinesModel( self.knots, order=self.order, copy=self )

    def calcNp( self, knots, order ) :
        if Tools.length( order ) == 1 :
            order = [order] * len( knots )
        np = 1
        for (kn,dr) in zip( knots, order ) :
            np *= ( len( kn ) + dr - 1 )

        if isinstance( np, numpy.int64 ) :
            np = np.item()

        return np

    def __setattr__( self, name, value ):
        """
        Set attributes: knots, order, models

        """
        dlst = {'order': int, 'models': Model }
        done = {'knots': list }
        if ( Tools.setListOfAttributes( self, name, value, dlst ) or
             Tools.setSingleAttributes( self, name, value, done ) ) :
            pass
        else :
            super( SurfaceSplinesModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)
        parlist : array_like
            not used in this model
        """
        ndata = Tools.length( xdata[:,0] )
        partial = numpy.ones( ( ndata, 1 ), dtype=float )
        np = 1
        n = 0
        for mdl in self.models :
            nw = numpy.zeros( ( ndata, np * mdl.npbase ), dtype= float )
            mpart = mdl.basePartial( xdata[:,n], params )
            k = 0
            for i in range( np ) :
                for j in range( mdl.npbase ) :
                    nw[:,k] = partial[:,i] * mpart[:,j]
                    k += 1

            partial = nw
            n += 1
            np *= mdl.npbase

        return partial

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        strkn = "["
        stror = "["
        for k,o in zip( self.knots, self.order ) :
            strkn += ( "%d,"%len( k ) )
            stror += ( "%d,"%o )
        strkn = strkn[:-1] + "]"
        stror = stror[:-1] + "]"
        return str( "SurfaceSplines %dd of order "%self.ndim + stror + " with " + strkn + " knots" )

    def baseParameterName( self, k ):
        """
        Return the name of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        strpar = "param"
        for mdl in self.models :
            nx = mdl.npbase
            strpar = strpar + "_%d"%(k%nx)
            k = k // nx
        return strpar

    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        u = self.yUnit
        n = 0
        for mdl in self.models :
            nx = mdl.npbase
            j = k % nx
            if j > 3 : j = 3
            k = k // nx
            u = u / ( self.xUnit[n] ** j )
            n += 1
        return u





import time
import numpy as numpy
from astropy import units
import math
from . import Tools
import matplotlib.pyplot as pyplot

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class IterationPlotter( object ):
    """
    The IterationPlotter plots intermediate results from a iterative fitter.
    <p>
    Author:      Do Kester

    """
    def plotData( self, x, y, title ):
        """
        Plot the data.

        Parameters :
            x       x-axis values of the data
            y       y-axis values of the data
            title   the title of the plot

        pyplot.figure( 1 )
        pyplot.plot( x, y, 'k.' )
        pyplot.title( title )
        pyplot.show( block=False )
        """

        self.p = pyplot.gca()
        self.p.plot( x, y, 'k.' )
        pyplot.show( block=False )
        time.sleep( 1 )

    def plotResult( self, x, r, iter ):
        """
        Plot the ( intermediate ) result.

        Parameters :
            x       x-axis values of the data
            r       model result
            iter    iteration number

        pyplot.figure( 1 )
        pyplot.plot( x, r, 'r-' )
        pyplot.show( block=False )
        """
        self.p.plot( x, r, 'r-' )
        pyplot.show( block=False )
        time.sleep( 1 )

    def plotProgress( self, percent ):
        """
        Plot ( estimated ) progress upto now.

        """
        pass





import numpy as numpy
from . import Tools
from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2017 Do Kester

class SplinesModel( LinearModel ):
    """
    General splines model of arbitrary order and with arbitrary knot settings.
    It is a linear model.

    order   behaviour between knots     continuity at knots
      0     piecewise constant          not continuous at all
      1     piecewise linear            lines are continuous
      2     parabolic pieces            1st derivatives are also continuous
      3     cubic pieces                2nd derivatives are also continuous
     n>3    n-th order polynomials      (n-1)-th derivatives are also continuous

    The user lays out a number ( << datapoints ) of knots on the x-axis at
    arbitrary position, generally more knots where the curvature is higher.
    The knots need to be monotonuously increasing in x.
    Alternatively one can ask this class to do the lay-out which is then
    equidistant in x over the user-provided range.
    Through these knots a splines function is obtained which best
    fits the datapoints. One needs at least 2 knots, one smaller and one
    larger than the x-values in the dataset.

    If the end knots are put in between the x-values in the dataset, a kind of
    extrapoling spline is obtained. It still works more or less. Dont push it.

    This model is NOT for (cubic) spline interpolation.

    Examples
    --------
    >>> knots = numpy.arange( 17, dtype=float ) * 10    # make equidistant knots from 0 to 160
    >>> csm = SplinesModel( knots=knots, order=2 )
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> csm = SplinesModel( nrknots=17, order=2, min=0, max=160 )    # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> npt = 161                                               # to include both 0 and 160.
    >>> x = numpy.arange( npt, dtype=float )                    # x-values
    >>> csm = SplinesModel( nrknots=17, order=2, xrange=x )     # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18


    Limitations
    -----------
    Dont put the knots too closely so that there are no datapoints in between.

    """
    def __init__( self, knots=None, order=3, nrknots=None, min=None, max=None, xrange=None,
                        copy=None, **kwargs ):
#                        copy=None, fixed=None, **kwargs ):
        """
        Splines on a given set of knots and a given order.

        The number of parameters is ( length( knots ) + order - 1 )

        Parameters
        ----------
        knots : array_like
            a array of arbitrarily positioned knots
        order : int
            order of the spline. Default 3 (cubic splines)
        nrknots : int
            number of knots, equidistantly posited over xrange or [min,max]
        min : float
            minimum of the knot range
        max : float
            maximum of the knot range
        xrange : array_like
            range of the xdata
        copy : SplinesModel
            model to be copied.
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        ValueError : At least either (`knots`) or (`nrnkots`, `min`, `max`) or
                (`nrknots`, `xrange`) must be provided to define a valid model.
        AttributeErrr : When fixed is not None

        Notes
        -----
        The SplinesModel is only strictly valid inside the domain defined by the
        minmax of knots. It deteriorates fastly going outside the domain.

        """
#        if fixed is not None :
#            raise AttributeError( "SplinesModel cannot have fixed parameters" )

        if copy is not None :
            knots = copy.knots
            order = copy.order
        if knots is not None : nrknots = len( knots )
        elif nrknots is None :
            raise ValueError( "Need either knots or (nrknots,min,max) or (nrknots,xrange)" )

        super( SplinesModel, self ).__init__( order + nrknots - 1, copy=copy, **kwargs )
        self.order = order
        if knots is None :
            if xrange is not None :
                min = numpy.min( xrange )
                max = numpy.max( xrange )
            knots = numpy.linspace( min, max, nrknots, dtype=float )
        self.knots = knots

    def copy( self ):
        return SplinesModel( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: knots, order

        """
        dlst = {'knots': float }
        dind = {'order': int }
        if ( Tools.setListOfAttributes( self, name, value, dlst ) or
             Tools.setSingleAttributes( self, name, value, dind ) ):
            pass
        else :
            super( SplinesModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npmax
        ni = Tools.length( xdata )
        partial = numpy.zeros( ( ni, np), dtype=float )
        x = numpy.ones( ni )
        for i in range( self.order ):
            partial[:,i] = x
            x *= xdata

        i = self.order
        ko = i - 1

        ki = numpy.arange( ni )
        while i < np - 1 :
            partial[ki,i] = x[ki]

            ki = numpy.where( xdata >= self.knots[i - ko] )[0]
            if ki.size == 0 : break

            d = xdata[ki] - self.knots[i-ko]
            t = numpy.ones( len( d ) )
            for k in range( self.order ):
                t *= d
            partial[ki,i] -= t

            x[ki] = t
            i += 1
        partial[ki,i] = x[ki]

        if parlist is None or len( parlist ) == np :
            return partial

        return partial[:,parlist]


    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model

        """
        ff = numpy.append( numpy.arange( self.order + 1, dtype=float ),
                           numpy.full( self.npbase - self.order - 1, 3, dtype=float ) )
        pars = params * ff
        pars = pars[1:]
        return SplinesModel( knots=self.knots, order=self.order-1 ).result( xdata, pars )


    def baseName( self ):
        """ Returns a string representation of the model. """
        return "Splines of order %d with %d knots."%( self.order, len( self.knots) )

    def baseParameterUnit( self, k ):
        """
        Return the name of the parameter.

        Parameters
        ----------
        k : int
            index of the parameter.
        """
        if k > self.order :
            k = self.order
        return self.yUnit / ( self.xUnit ** k )





import numpy as numpy
from . import Tools

from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class StepEngine( Engine ):
    """
    Move a a walker in a random direction.

    The StepEngine tries to move a selection of the parameters
    in a random order.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : StepEngine
            to be copied
        seed : int
            for rng

        """
        super( StepEngine, self ).__init__( walkers, errdis, copy=copy, seed=seed )

    def copy( self ):
        """ Return copy of this.  """
        return StepEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StepEngine" )


    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        walker : Sample
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of the/some parameters indices to be diffused

        Returns
        -------
        int : the number of successfull moves

        """
        if fitIndex is None :
            fitIndex = walker.fitIndex
        np = len( fitIndex )

        model = walker.model
        nm = len( self.walkers )

        urange = self.unitRange[fitIndex]
        dur = urange / nm
        urange += 2 * dur

        param = walker.allpars
        usav = self.domain2Unit( model, param, kpar=fitIndex )

        sz = 1.0
        ptry = param.copy()
        kk = 0
        while True :
            kk += 1

            step = ( 2 * self.rng.rand( np ) - 1 ) * urange
            while True :
                utry = usav + step
                q0 = numpy.where( utry < 0 )[0]
                nq0 = len( q0 )
                q1 = numpy.where( utry > 1 )[0]
                nq1 = len( q1 )

                if nq0 > 0 :
                    step[q0] = ( 2 * self.rng.rand( nq0 ) - 1 ) * urange[q0]
                elif nq1 > 0 :
                    step[q1] = ( 2 * self.rng.rand( nq1 ) - 1 ) * urange[q1]
                else :
                    break

            ptry[fitIndex] = self.unit2Domain( model, utry,  )

            Ltry = self.errdis.logLikelihood( model, ptry )
            if Ltry >= lowLhood:
                self.reportSuccess( )
                self.setSample( walker, model, ptry, Ltry )
                break
            elif kk < self.maxtrials :
                sz *= 0.5
                self.reportReject( )
            else :
                self.reportFailed()
                break

        return np                        # nr of succesfull steps





import numpy as numpy
from astropy import units
import math
from . import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class Engine( object ):
    """
    Engine defines general properties of a Engine.

    An Engine moves around a sample in a random way such that its likelood
    remain above the low-likelihood-limit.

    Attributes
    ----------
    walkers : SampleList
        walkers to be diffused
    errdis : ErrorDistribution
        error distribution to be used
    maxtrials : int
        maximum number of trials for various operations
    rng : numpy.random.RandomState
        random number generator

    report : list of int (read only)
        reports number of succes, accepted, rejected, failed calls. Plus the total.
    unitRange : array_like (read only)
        present max size of the parameter cloud (in [0,1])
    unitMin : array_like (read only)
        present minimum values of the parameter cloud (in [0,1])

    Author       Do Kester.

    """
    SUCCESS = 0             #  succesfull move
    REJECT = 1              #  rejected move
    FAILED = 2              #  failed to move
    NCALLS = 3              #  number of calls


    #  *********CONSTRUCTORS***************************************************

    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Copy Constructor.
        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : Engine
            engine to be copied

        """
        self.walkers = walkers
        self.errdis = errdis
        self.report = [0]*4

        if copy is None :
            self.maxtrials = 5
            self.rng = numpy.random.RandomState( seed )
        else :
            self.maxtrials = copy.maxtrials
            self.rng = copy.rng

    def copy( self ):
        """ Return a copy of this engine.  """
        return Engine( self.walkers, self.errdis, copy=self )

    def __getattr__( self, name ) :
        if name == "unitRange" :
            return self.calculateUnitRange()
        elif name == "unitMin" :
            self.calculateUnitRange()
            return self.unitMin
        else :
             raise AttributeError( str( self ) + ": Unknown attribute " + name )


    #  *********SET & GET***************************************************
    def setSample( self, walker, model, allpars, logL, logW=None,
                   fitindex=None ):
        """
        Update the sample with model, allpars, LogL and logW.

        Parameters
        ----------
        walker : Sample
            sample to be updated
        model : Model
            the model in the sample
        allpars : array_like
            list of all parameters
        logL : float
            log Likelihood
        logW : float
            log of the weight of the likelihood
        fitindex : array_like
            list of parameter indices to be diffuse
        """
        walker.logL = logL
        walker.allpars = allpars
        walker.model = model
        if logW is not None :
            walker.logW = logW
        if fitindex is not None :
            walker.fitIndex = fitindex
        self.walkers[walker.id] = walker

###### TBD:
##  Has dval the length kpar ???

    def domain2Unit( self, model, dval, kpar=None ) :
        """
        Return value in [0,1] for the selected parameter.

        Parameters
        ----------
        model : Model
            the model involved
        dval : float
            domain value for the selected parameter
        kpar : None or array_like
            selected parameter index, where kp is index in [parameters, hyperparams]
            None means all
        """
        np = model.npchain
        if kpar is None :
            kpar = self.makeIndex( np, dval )

        elif Tools.isInstance( kpar, int ) :
            return ( model.domain2Unit( dval, kpar ) if kpar >= 0 else
                     self.errdis.domain2Unit( dval, kpar ) )
#           return ( model.domain2Unit( dval, kpar ) if kpar < np else
#                     self.errdis.domain2Unit( dval, kpar - np ) )

        uval = numpy.ndarray( len( kpar ), dtype=float )


#        print( "Eng d2u  ", dval, kpar, uval )
        for i,kp in enumerate( kpar ) :
#            if kp < np :
            if kp >= 0 :
                uval[i] = model.domain2Unit( dval[kp], kp )
            else :
#                uval[i] = self.errdis.domain2Unit( dval[kp], kp - np )
                uval[i] = self.errdis.domain2Unit( dval[kp], kp )
        return uval

    def unit2Domain( self, model, uval, kpar=None ) :
        """
        Return domain value for the selected parameter.

        Parameters
        ----------
        model : Model
            the model involved
        uval : array_like
            unit value for the selected parameter
        kpar : None or array_like
            selected parameter indices, where kp is index in [parameters, hyperparams]
            None means all.
        """
        np = model.npchain

        if kpar is None :
            kpar = self.makeIndex( np, uval )
        elif Tools.isInstance( kpar, int ) :
            return ( model.unit2Domain( uval, kpar ) if kpar >= 0 else
                     self.errdis.unit2Domain( uval, kpar ) )
#            return ( model.unit2Domain( uval, kpar ) if kpar < np else
#                     self.errdis.unit2Domain( uval, kpar - np ) )

        dval = numpy.ndarray( len( kpar ), dtype=float )
        for i,kp in enumerate( kpar ) :
#            if kp < np :
            if kp >= 0 :
                dval[i] = model.unit2Domain( uval[kp], kp )
            else :
#                dval[i] = self.errdis.unit2Domain( uval[kp], kp - np )
                dval[i] = self.errdis.unit2Domain( uval[kp], kp )
        return dval

    def makeIndex( self, np, val ) :
        kpar = [k for k in range( np )]
        nh = len( val ) - np
        kpar += [-k for k in range( nh, 0, -1 )]
        return kpar

    def reportCall( self ):
        """ Store a call to engine  """
        self.report[self.NCALLS] += 1

    def reportSuccess( self ):
        self.report[self.SUCCESS] += 1

    def reportReject( self ):
        self.report[self.REJECT] += 1

    def reportFailed( self ):
        self.report[self.FAILED] += 1

    def printReport( self ) :
        print( " %10d %10d %10d %10d" % (self.report[0], self.report[1],
                                         self.report[2], self.report[3] ) )

    def calculateUnitRange( self ):
        """
        Calculate the range of the present parameter values in unit values.

        For Dynamic models the range is calculated for those parameters present in all models;
        it is 1.0 for other parameters.

        """
        kmx = 0
        if not self.walkers[0].model.isDynamic() :
#            npmax = len( self.walkers[0].allpars )
            npmax = len( self.walkers[0].fitIndex )
        else :
            npmax = 0
            for k, walker in enumerate( self.walkers ) :
                if len( walker.allpars ) > npmax :
#                    npmax = len( walker.allpars )
                    npmax = len( walker.fitIndex )
                    kmx = k
#        minv = self.walkers[kmx].allpars.copy()
        minv = numpy.ones( npmax, dtype=float )
        maxv = numpy.zeros( npmax, dtype=float )
        nval = numpy.zeros( npmax, dtype=int )

#        print( "npmax  ", npmax, nval )

        for walker in self.walkers :
#            print( "ENG   ", walker.id, walker.fitIndex, walker.allpars )
            fi = walker.fitIndex
            minv[fi] = numpy.fmin( minv[fi], walker.allpars[fi] )
            maxv[fi] = numpy.fmax( maxv[fi], walker.allpars[fi] )
            nval[fi] += 1

        model = self.walkers[kmx].model

        maxv = self.domain2Unit( model, maxv, kpar=self.walkers[kmx].fitIndex )
        minv = self.domain2Unit( model, minv, kpar=self.walkers[kmx].fitIndex )

#        print( maxv <= minv, nval < len( self.walkers ) )
        q = numpy.where( numpy.logical_or( maxv <= minv, nval < len( self.walkers ) ) )
        maxv[q] = 1.0
        minv[q] = 0.0

        self.unitRange = numpy.abs( maxv - minv )
        self.unitMin = numpy.fmin( minv, maxv )

        return self.unitRange

    def __str__( self ) :
        return str( "Engine" )

    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine for diffusing the parameters

        Parameters
        ----------
        walker : Sample
            walker to diffuse
        lowLhood : float
            low limit on the loglikelihood
        fitIndex : array_like
            list of parameter indices to diffuse

        Returns
        -------
        int : number of succesfull moves

        """
        pass


class DummyPlotter( object ) :

    def __init__( self, iter=0 ) :
        self.iter = iter

    def start( self ):
        """ start the plot. """
        pass

    def move( self, param, ptry, col=None ):
        """
        Move parameters at position param to ptry using color col.
        """
        pass

    def stop( self ):
        """ Stop (show) the plot. """
        pass




#
#  This file is part of the BayesicFitting package.
#
#  BayesicFitting is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation, either version 3 of
#  the License, or (at your option) any later version.
#
#  BayesicFitting is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#


from .Biweight import Biweight
from .CosSquare import CosSquare
from .Cosine import Cosine
from .Gauss import Gauss
from .Huber import Huber
from .Kernel import Kernel
from .Lorentz import Lorentz
from .Parabola import Parabola
from .Sinc import Sinc
from .Triangle import Triangle
from .Tricube import Tricube
from .Triweight import Triweight
from .Uniform import Uniform




import numpy as numpy
import math
from . import Tools
from .NonLinearModel import NonLinearModel
from .PolynomialModel import PolynomialModel
from .SplinesModel import SplinesModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester

class SineSplineDriftModel( NonLinearModel ):
    """
    Sine with drifting frequency and splineslike amplitudes/phases.

    .. math::
        nd = degree + 1
        nh = len( knots ) + order - 1
        xx = 2 \pi x PM( x:p[:nd] )
        A  = SM( x:p[nd:nd+nh] )
        B  = SM( x:p[nd+nh:] )
        f( x:p ) = A \cos( xx ) + B \sin( xx )

    Where :math:`s_0` and :math:`s_1` are splines with defined knots and order.
    It is a linear model with 2 * ( len(knots) + order - 1 ) papameters.

    Examples
    --------
    >>> knots = [3.0*k for k in range( 11 )]
    >>> sine = SineSplineDriftModel( 150, knots )        # fixed frequency of 150 Hz
    >>> print( sine.npbase )                        # number of parameters
    26
    """

    def __init__( self, knots, order=3, degree=1, copy=None, fixed=None, **kwargs ):
        """
        Sine model of a fixed frequency with a splineslike changing
        amplitude/phase and polynomially changing frequency.

        Number of parameters is 2 * ( len(knots) + order - 1 ) + degree + 1.

        Parameters
        ----------
        frequency : float
            the frequency
        copy : SineSplineDriftModel
            model to be copied
        fixed : dict
            If not None raise AttributeError.

        Raises
        ------
        AttributeError : When fixed is not None

        """
        if fixed is not None :
            raise AttributeError( "FreeShapeModel cannot have fixed parameters" )

        np = 2 * ( len( knots ) + order - 1 ) + degree + 1
        super( SineSplineDriftModel, self ).__init__( np, copy=copy, **kwargs )

        self.knots = knots
        self.order = order
        self.degree = degree
        if copy is not None :
            self.pm = copy.pm.copy()
            self.cm = copy.cm.copy()
            self.sm = copy.sm.copy()
        else :
            self.pm = PolynomialModel( self.degree )
            self.cm = SplinesModel( knots, order=order )
            self.sm = SplinesModel( knots, order=order )

    def copy( self ):
        """ Copy method.  """
        return SineSplineDriftModel( self.knots, order=self.order,
                                     degree=self.degree, copy=self )

    def __setattr__( self, name, value ) :
        dind = {"degree": int, "order": float, "pm" : PolynomialModel,
                "cm": SplinesModel, "sm": SplinesModel}
        dlst = {"knots": float}
        if not ( Tools.setListOfAttributes( self, name, value, dlst ) or
                 Tools.setSingleAttributes( self, name, value, dind ) ) :
            super( SineSplineDriftModel, self ).__setattr__( name, value )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        drift = self.pm.result( xdata, params[:self.degree+1] )
        x = 2 * math.pi * xdata * drift
        amps = self.getAmplitudes( xdata, params )
        return amps[0] * numpy.cos( x ) + amps[1] * numpy.sin( x )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model (ignored in LinearModels)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        nxdata = Tools.length( xdata )
        part = numpy.zeros( ( nxdata, self.npbase ), dtype=float )
        nd = self.degree + 1
        nh = ( self.npbase - nd ) // 2
        drift = 2 * math.pi * xdata * self.pm.result( xdata, params[:nd] )
        amps = self.getAmplitudes( xdata, params )
        cx = numpy.cos( drift )
        sx = numpy.sin( drift )
        dpm = 2 * math.pi * xdata * self.pm.basePartial( xdata, params ).transpose()
        part[:,:nd]      = ( ( amps[1] * cx * dpm ).transpose() -
                             ( amps[0] * sx * dpm ).transpose() )
        part[:,nd:nd+nh] = ( cx * self.cm.basePartial( xdata, params ).transpose() ).transpose()
        part[:,nd+nh:]   = ( sx * self.sm.basePartial( xdata, params ).transpose() ).transpose()
        return part

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        nd = self.degree + 1
        drift = 2 * math.pi * xdata * self.pm.result( xdata, params[:nd] )
        cx = numpy.cos( drift )
        sx = numpy.sin( drift )
        pdx = self.pm.baseDerivative( xdata, params )
        cdx = cx * pdx
        sdx = sx * pdx
        amps = self.getAmplitudes( xdata, params )
        cadx = self.cm.baseDerivative( xdata, params )
        sadx = self.sm.baseDerivative( xdata, params )
        return 2 * math.pi * ( cadx * cx - amps[0] * sdx + sadx * sx + amps[1] * cdx )

    def getAmplitudes( self, xdata, params ) :
        """
        Return the amplitudes if cosine and sine, resp.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters of the model

        """
        nd = self.degree + 1
        nh = ( self.npbase - nd ) // 2
        return ( self.cm.result( xdata, params[nd:nd+nh] ),
                 self.sm.result( xdata, params[nd+nh:] ) )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return ( "SineSplineDrift: f( x:p ) = spline_0 * cos( poly ) + " +
                 "spline_1 * sin( poly )" )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        k = k % ( self.npbase / 2 )
        if k > self.order :
            k = self.order
        return self.yUnit / ( self.xUnit ** k )






import numpy as numpy
import math

from .ErrorDistribution import ErrorDistribution
from .LogFactorial import logFactorial

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class PoissonErrorDistribution( ErrorDistribution ):
    """
    To calculate a Poisson likelihood.

    For one observation with n counts it holds
    ..math ::
        f( n,x ) = x^n / ( e^x * n! )

    where x is the expected counts

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.
    .. math::
        logL = \sum( n * \log( x ) - x - \log( n! ) )

    Author       Do Kester.

    """

    PARNAMES = []

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        xdata : array_like
            input data for the model
        data : array_like
            data to be fitted
        weights : array_like
            weights to be used (no weights are possible in Poisson)
        copy : PoissonErrorDistribution
            distribution to be copied.

        Raises
        ------
        ValueError when weights is not None

        """
        if weights is not None:
            raise ValueError( "Weights are not possible in Poisson distributions" )

        super( PoissonErrorDistribution, self ).__init__( xdata, data, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return PoissonErrorDistribution( self.xdata, self.data, copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always false for this distribution.
        """
        return False

    def getScale( self, model ) :
        """
        Return the sqrt( sum( data ) / N )

        """
        return math.sqrt( sum( self.data ) / self.sumweight )


    #  *********LIKELIHOODS***************************************************
    def logLikelihood( self, model, param ):
        """
        Return the log( likelihood ) for a Poisson distribution.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        param : array_like
            parameters of the model

        """
        self.ncalls += 1
        mock = model.result( self.xdata, param )
        if numpy.any( numpy.less_equal( mock, 0 ) ) :
            return -math.inf

        lfdata = logFactorial( self.data )

        logl = numpy.sum( self.data * numpy.log( mock ) - mock - lfdata )

        if math.isnan( logl ) :
            return -math.inf

        return logl

    def partialLogL( self, model, param, fitIndex ):
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        param : array_like
            parameters of the model
        fitIndex : array_like
            indices of the params to be fitted
        """
        self.nparts += 1
        mock = model.result( self.xdata, param )
        dM = model.partial( self.xdata, param )
        dL = numpy.zeros( len( fitIndex ), dtype=float )

        i = 0
        for k in fitIndex :
            dL[i] = numpy.sum( ( self.data / mock - 1 ) * dM[:,k] )
            i += 1

        return dL

    def __str__( self ) :
        return "Poisson error distribution"




import numpy as numpy
from astropy import units
import math
from . import Tools

from .Model import Model
from .Dynamic import Dynamic
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018        Do Kester

class RepeatingModel( Model, Dynamic ):
    """
    RepeatingModel implements the the Dynamic interface for a Model.

    It repeatedly calls a  Model, called the basic model,
    zero ( or more ) times, each time with the next set of parameters.


    Author       Do Kester

    """
    #  *****CONSTRUCTOR*********************************************************
    def __init__( self, ncomp, model, minComp=0, maxComp=None, fixed=None,
                  growPrior=None, copy=None, **kwargs ):
        """
        Repeating the same model several times.

        Parameters
        ----------
        ncomp : int
            number of repetitions
        model : Model
            model to be repeated
        minComp : int (0)
            minimum number of repetitions
        maxComp : None or int
            maximum number of repetitions
        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior
        copy : HarmonicDynamicModel
            model to copy

        Raises
        ------
        AttributeError when fixed parameters are requested
        ValueError when order is outside [min..max] range

        """
        if fixed is not None :
            raise AttributeError( "DynamicModel cannot have fixed parameters" )
        if ncomp < minComp or ( maxComp is not None and ncomp > maxComp ) :
            raise ValueError( "ncomp outside range of [min..max] range" )

        np = ncomp * model.npchain
        super( RepeatingModel, self ).__init__( np, copy=copy, **kwargs )

        self.ncomp = ncomp
        self.model = model

        if copy is None :
            self.minComp = minComp
            self.maxComp = maxComp
            if growPrior is None :
                if maxComp is None :
                    self.growPrior = ExponentialPrior( scale=2 )
                else :
                    lim = [minComp, maxComp+1]        # limits on components
                    self.growPrior = UniformPrior( limits=lim )
            else :
                self.growPrior = growPrior
        else :
            self.minComp = copy.minComp
            self.maxComp = copy.maxComp
            self.growPrior = copy.growPrior.copy()

    def copy( self ):
        """ Copy method.  """
        return RepeatingModel( self.ncomp, self.model, copy=self )

    def changeNComp( self, dn ) :
        self.ncomp += dn

    #  *************************************************************************
    def __setattr__( self, name, value ) :
        dind = {"minComp": int, "maxComp": int, "growPrior": Prior,
                "model": Model, "ncomp" : int, "deltaNpar" : int}
        lnon = {"maxComp": int}

        if name == "model" :
            self.deltaNpar = value.npchain

        if ( Tools.setNoneAttributes( self, name, value, lnon ) or
             Tools.setSingleAttributes( self, name, value, dind ) ) :
            pass
        else :
            super( RepeatingModel, self ).__setattr__( name, value )

    #  *************************************************************************
    def result( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        result = numpy.zeros( Tools.length( xdata ), dtype=float )

        ks = 0
        for k in range( self.ncomp ) :
            ke = ks + self.deltaNpar
            result += self.model.result( xdata, params[ks:ke] )
            ks = ke

        return result

    def partial( self, xdata, params ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters


        """
        nxdata = Tools.length( xdata )
        partial = numpy.ndarray( ( nxdata, self.npbase ), dtype=float )

        ks = 0
        for k in range( self.ncomp ) :
            ke = ks + self.deltaNpar
            partial[:,ks:ke] = self.model.partial( xdata, params[ks:ke] )
            ks = ke
        return partial

    def derivative( self, xdata, params ):
        """
        Returns the derivative df/dx at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        nxdata = Tools.length( xdata )
        dfdx = numpy.ndarray( ( nxdata, self.ndim ), dtype=float )

        ks = 0
        for k in range( self.ncomp ) :
            ke = ks + self.deltaNpar
            dfdx += self.model.derivative( xdata, params[ks:ke] )
            ks = ke
        return dfdx

    def getNumberOfComponents( self ):
        """ Return the number of basic models inside this compound model.  """
        return self.ncomp


    def getPrior( self, k ):
        return super( RepeatingModel, self ).getPrior( k % self.deltaNpar )

    def baseName( self ):
        """ Return a string representation of the model.  """
        return ( "Repeat %d times:\n  " % self.ncomp +
                  self.model.baseName( ) )

    def getParName( self, k ):
        """
        Return the name of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        return ( self._model.getParameterName( k % self.deltaNpar ) + "_" +
                ( k / self.deltaNpar ) )

    def getParUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        self._model.setUnits( self.xUnit, self.yUnit )
        return self.model.getParameterUnit( k % self.deltaNpar )



import numpy as numpy
from astropy import units
import math
from . import Tools

from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class HarmonicModel( LinearModel ):
    """
    Harmonic oscillator Model.

    f( x:p ) = SUM_j ( p_k * cos( 2*pi*j*x ) + p_k+1 * sin( 2*pi*j*x ) )
    j = 1, N; k = 0, 2N-2, 2

    The number of parameters is 2 * order.
    The parameters are initialized at 1.0. It is a linear model.

    See
    <a href="../../../../../../ia/numeric/toolbox/fit/demo/harmonicfit.py">example</a>

    Author:      Do Kester

    Examples
    --------
    >>> harm = HarmonicModel( 3 )            # period = 1
    >>> print( harm.npbase
    6
    >>> harm = HarmonicModel( 4, 2.7 )        # period = 2.7

    Attributes
    ----------
    order : int
        the order of the harmonic
    period : float
        the length of the period of the fundamental

    """
    PARNAMES = ["cosamp_", "sinamp_"]

    def __init__( self, order, period=1.0, copy=None, **kwargs ):
        """
        Harmonic oscillator model.

        Number of parameters is 2 * order.

        Parameters
        ----------
        order : int (>0)
            the number of overtones
        period : float
            length of the period of the fundamental. default 1.0
        copy : HarmonicModel
            model to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        super( HarmonicModel, self ).__init__( 2 * order, copy=copy, **kwargs )

        self.order = order
        self.period = period

    def copy( self ):
        """ Copy method.  """
        return HarmonicModel( self.order, period=self.period, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: order, period

        """
        dind = {'order': int, 'period': float }
        if Tools.setSingleAttributes( self, name, value, dind ):
            pass
        else :
            super( HarmonicModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            x values at which to calculate the partials
        params : array_like
            parameters of the model. (ignored in LinearModels)
        parlist : array_like
            list of indices of active parameters

        """
        np = self.npmax if parlist is None else len( parlist )
        ni = Tools.length( xdata )
        partial = numpy.zeros( ( ni, np), dtype=float )

        x = 2 * math.pi * xdata / self.period

        parts = { 0 : ( lambda j : numpy.cos( j * x ) ),
                  1 : ( lambda j : numpy.sin( j * x ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            j = 1 + kp // 2
            partial[:,k] = parts[kp % 2]( j )

        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters of the model.

        """
        df = numpy.zeros_like( xdata )
        tp = 2 * math.pi / self.period
        x = tp * xdata
        j = 0
        for k in range( 0, self.npmax, 2 ) :
            j += 1
            df += tp * j * ( params[k+1] * numpy.cos( j * x ) - params[k] * numpy.sin( j * x ) )
        return df

    def baseName( self ):
        """ Returns a string representation of the model.  """
        bn = "Harmonic: f( x:p ) = "
        if self.order == 0 :
            bn += "0"
            return bn

        bn += "p_0 * cos( xx ) + p_1 * sin( xx )"
        j = 2
        for k in range( 2, self.order + 1 ) :
            bn += " + p_%d * cos( %d xx ) + p_%d * sin( %d xx )"%(j,k,j+1,k)
            j += 2
        bn += "; xx = 2 * pi * x / period"
        return bn

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return "%s%d"%( self.PARNAMES[k%2], k / 2 )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Always : YUnit.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return self.yUnit





import numpy as numpy

from .Kernel import Kernel
from .Cosine import Cosine

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  *    This file is part of the BayesicFitting package.
#  *
#  *    BayesicFitting is free software: you can redistribute it and/or modify
#  *    it under the terms of the GNU Lesser General Public License as
#  *    published by the Free Software Foundation, either version 3 of
#  *    the License, or ( at your option ) any later version.
#  *
#  *    BayesicFitting is distributed in the hope that it will be useful,
#  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  *    GNU Lesser General Public License for more details.
#  *
#  *    The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2017 Do Kester

class CosSquare( Kernel ):
    """
    CosSquare (Cosine Squared) is a Kernel function between [-1,1]; it is 0 elsewhere.
    .. math::
        K( x ) = cos^2( 0.5*PI*x ) if |x| < 1 else  0


    """
    def __init__( self ) :
        super( CosSquare, self ).__init__( integral=1.0, fwhm=1.0 )

    def result( self, x ):
        res = Cosine.result( Cosine, x )
        return res * res

    def resultsq( self, xsq ):
        return self.result( numpy.sqrt( xsq ) )

    def partial( self, x ):
        return 2 * Cosine.result( Cosine, x ) * Cosine.partial( Cosine, x )

    def isBound( self ):
        return True

    def name( self ):
        return str( "CosSquare: cos^2( 0.5*PI*x ) if |x| < 1 else 0" )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Formatter import formatter as fmt

from .Dynamic import Dynamic
from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018        Do Kester

class BirthEngine( Engine ):
    """
    The BirthEngine adds a new component to the model.

    Only for Models that are Dynamic.
    The birth rate is governed by the growth-prior in the dynamic model.

    The member is kept when the logLikelihood > lowLhood.

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=23455 ) :
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : GalileanEngine
            to be copied
        seed : int
            for random number generator

        """
        super( BirthEngine, self ).__init__( walkers, errdis, copy=copy,
                    seed=seed )

    def copy( self ):
        """ Return copy of this.  """
        return BirthEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "BirthEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by adding a component and diffusing the parameters.

        Parameters
        ----------
        walker : Sample
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of the/some parameters indices to be diffused

        Returns
        -------
        int : the number of successfull moves

        """
        waltry = walker.copy()

        model = waltry.model
        ptry = waltry.allpars

        pat = 0
        while model is not None and not isinstance( model, Dynamic ) :
            pat += model.npbase
            model = model._next

        nc = model.getNumberOfComponents()
        np = model.npbase

        if not ( nc < model.growPrior.unit2Domain( self.rng.rand() ) and
                 model.grow( pat ) ):
            self.reportFailed()
            walker.check( nhyp=self.errdis.nphypar )
            return 0

        dnp = model.deltaNpar        # parameter increase
        ptry = model.alterParameters( ptry, np, dnp, pat )
        find = model.alterFitindex( walker.fitIndex, np, dnp, pat )

        model = waltry.model
        t = 0
        while t < self.maxtrials :
            for kp in range( pat+np, pat+np+dnp ) :
                ptry[kp] = model.unit2Domain( self.rng.rand(), kpar=kp )
            Ltry = self.errdis.logLikelihood( model, ptry )

            if Ltry >= lowLhood:
                self.reportSuccess()
                self.setSample( walker, model, ptry, Ltry, fitindex=find )
                walker.check( nhyp=self.errdis.nphypar )
                return dnp
            else :
                self.reportFailed()
                t += 1

        self.reportReject( )
        walker.check( nhyp=self.errdis.nphypar )

        return 0





import numpy as numpy
from astropy import units
import math
from . import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2006 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class ImageAssistant( object ):
    """
    ImageAssistant contains 2 methods to assist with more dimensional
    fitting.

    1. getIndices Generates indices for data arrays of any dimension.
       To be used as input in the Fitter classes.
    2. resizeData Resizes the data arrays into a 1-dimensional array.
       To be used as data in the Fitter.




    Example
    -------
    >>> ymap = numpy.arange( 6, dtype=float ).reshape( 2, 3 )
    >>> ias = ImageAssistant()
    >>> ky = ias.getIndices( ymap )
    >>> print( ky.shape )
        (6,2)
    >>> print( ky[4,0], ky[4,1], ymap[ ky[4,0], ky[4,1] ] )
        1 0 4
    >>> ias = ImageAssistant( order='F')
    >>> ky = ias.getIndices( ymap )
    >>> print( ky.shape )
        (6,2)
    >>> print( ky[4,0], ky[4,1], ymap[ ky[4,1], ky[4,0] ] )
        0 1 4

    ## Suppose y is a 2-dimensional map of something
    >>> aass = ImageAssistant( )
    >>> input = aass.getIndices( y )
    >>> fitter = Fitter( input, some2dModel )
    >>> pars = fitter.fit( aass.resizeData( y ) )
    >>> yfit = some2dModel.result( input )        # Double1d
    >>> yfit2d = aass.resizeData( yfit, y )        # Double2d


    Author       Do Kester

    """


    def __init__( self, order='C' ):
        """
        Helper class to construct the input arrays needed for the Fitters.

        Parameters
        ----------
        order : 'C' or 'F'
            set index view according to character
            'C' orders from slow to fast
            'F' orders from fast to slow
        """
        self.order = order
        self.shape = None

    def getIndices( self, ya, order='C' ):
        """
        Generates indices for data arrays of any dimension.

        To be used as input in the Fitter classes.

        Parameters
        ----------
        ya : map
            array of y ( data ) values for which an indexed array
        order : 'C' or 'F'
            set index view according to character

        Returns
        -------
        numpy.array of ints : the indices of the pixels

        """
        self.shape = ya.shape

        size = ya.size
        rank = ya.ndim

        if rank == 1:
            return numpy.arange( size, dtype=int )
        else:
            kdata = numpy.zeros( ( size, rank ), dtype=int )

            dd = 1
            kr = rank - 1
            for k in range( rank ) :
                for i in range( size ) :
                    kk = kr if self.order == 'C' else k
                    kdata[i,kk] = ( i // dd ) % self.shape[kr]
                dd *= self.shape[kr]
                kr -= 1

            return kdata

    def getPositions( self, ymap, order='C', center=True, deproject=None ) :
        """
        Return the (x,y) positions of the pixels in the map.

        Parameters
        ----------
        ya : map
            array of y ( data ) values for which an indexed array
        order : 'C' or 'F'
            set index view according to character
        center : bool
            if True, return the positions of the center of the pixels.
            otherwise the (left,lower) corner
        deproject : callable
            Deprojection method: from projected map to sky position,
            returning (x,y,...) position given the map indices (ix,iy,...)
            Default: returning the indices as floats (+0.5 if center)

        Returns
        -------
        numpy.array of floats : the positions of the pixels
        """
        xdata = numpy.asarray( self.getIndices( ymap, order=order ), dtype=float )
        if center :
            xdata += 0.5
        if deproject is not None :
            xdata = deproject( xdata )
        return xdata


    def getydata( self, ya ):
        """
        Return a copy of ya as a 1 dim array.

        Parameters
        ----------
        ya : array_like
            map to be reshaped
        """
        return numpy.reshape( ya.copy(), ya.size, order=self.order )

    def resizeData( self, res, shape=None ):
        """
        Reshape the data (res) into the same shape as the map (ya)

        Parameters
        ----------
        res : array_like
            result of the fit as a 1-dim array
        shape : tuple of int
            dimensional lengths of the reconstructable map
            default remembered from a call to getIndices
        """
        if shape is None :
            shape = self.shape

        return numpy.reshape( res, shape, order=self.order )



